//@ts-check
"use strict";

/**
@typedef Baxton
@type {{
  pos: PIXI.Point;
  armAngle: number;
  collideBox: { width: number; height: number; };
  hitbox: { width: number; height: number; };
  shootCooldown: number;
  speed: number;

  hitKnockback: PIXI.Point;
  hitTimer: number;

  immuneTimer: number;

  pixiRoot: PIXI.Container;
  bodySprite: PIXI.AnimatedSprite;
  bodySpritesheet: Spritesheet;
  armContainer: PIXI.Container;
  armSprite: PIXI.Sprite;

  shadowSprite: PIXI.Sprite;
  footstepSfx: PIXI.sound.IMediaInstance;
  footstepFxTimer: number;

  area: Area;
}}
 */

const playerMetrics = (() => {
  let heroAnchor = new PIXI.Point(.5, .8);
  let heroSize = new PIXI.Point(64, 64);
  let center = new PIXI.Point(
    heroSize.x * heroAnchor.x,
    heroSize.y * heroAnchor.y,
  );
  let armSocketPos = new PIXI.Point(40, 37);
  let centerToArmSocket = new PIXI.Point(
    armSocketPos.x - center.x,
    armSocketPos.y - center.y,
  );

  let canonHolePos = new PIXI.Point(55, 25);
  let armCenter = new PIXI.Point(32, 32);
  let socketToHole = new PIXI.Point(
    canonHolePos.x - armCenter.x,
    canonHolePos.y - armCenter.y,
  );

  return {
    heroAnchor, heroSize, center, armSocketPos, centerToArmSocket, socketToHole,
  };
})();

/** @returns {Baxton} */
function createBaxton() {
  const pixiRoot = new PIXI.Container();
  pixiRoot.scale.set(inverseTileSize);

  const armContainer = new PIXI.Container();
  armContainer.position.copyFrom(playerMetrics.centerToArmSocket);
  pixiRoot.addChild(armContainer);

  // arm sprites are aligned: they connect with the arm socket in the center
  const armSprite = new PIXI.Sprite(tex("inGame/baxton/baxtonArmFiring"));
  armSprite.anchor.set(0.5, 0.5);
  armContainer.addChild(armSprite);

  const bodySpritesheet = getSpritesheet("inGame/baxton/baxtonBody");
  const bodySprite = new PIXI.AnimatedSprite(bodySpritesheet.animations["idle"], false);
  bodySprite.anchor.copyFrom(playerMetrics.heroAnchor);
  pixiRoot.addChild(bodySprite);

  const shadowSprite = new PIXI.Sprite(tex("inGame/shadow"));
  shadowSprite.scale.set(inverseTileSize * 1.2);
  shadowSprite.anchor.set(0.5, 0);

  /** @type {Baxton} */
  const baxton = {
    pos: new PIXI.Point(0, 0),
    armAngle: 0,
    collideBox: { width: 0.8, height: 0.45 },
    hitbox: { width: 0.9, height: 1.4 },
    shootCooldown: 0,
    speed: 5,

    hitKnockback: new PIXI.Point(),
    hitTimer: 0,

    immuneTimer: 0,

    pixiRoot,
    bodySprite,
    bodySpritesheet,
    armContainer,
    armSprite,

    shadowSprite,
    footstepSfx: null,
    footstepFxTimer: 0,

    area: null
  };

  return baxton;
}

/** @param {number} dt - Delta time in seconds */
function updateBaxton(dt) {
  const { target, baxton } = inGame;

  if (baxton.area == null || !baxton.area.rect.contains(baxton.pos.x, baxton.pos.y)) setupBaxtonArea(true);

  function updateRenderPos() {
    baxton.pixiRoot.position.copyFrom(baxton.pos);
    baxton.pixiRoot.zIndex = baxton.pos.y;
    baxton.shadowSprite.position.copyFrom(baxton.pos);
    baxton.shadowSprite.zIndex = baxton.pos.y - 10;
  }

  if (inGame.gameOver) {
    baxton.bodySprite.update(app.ticker.deltaTime);
    return;
  }

  if (baxton.hitTimer > 0) {
    baxton.hitTimer -= dt;
    baxton.bodySprite.update(app.ticker.deltaTime);

    moveWithCollision(baxton, baxton.hitKnockback);
    updateRenderPos();

    baxton.hitKnockback.x *= 0.9;
    baxton.hitKnockback.y *= 0.9;

    if (baxton.hitTimer <= 0) baxton.bodySprite.loop = true;
    return;
  }

  if (baxton.immuneTimer > 0) baxton.immuneTimer -= dt;

  // Movement
  const speed = dt * baxton.speed;
  const direction = { x: 0, y: 0 };

  const { mouse, actions } = inGame.input;
  if (actions.moveUp) direction.y -= 1;
  if (actions.moveDown) direction.y += 1;
  if (actions.moveLeft) direction.x -= 1;
  if (actions.moveRight) direction.x += 1;

  let oldPos = baxton.pos.clone();
  let centerToArmSocket = playerMetrics.centerToArmSocket.clone();

  if (!inGame.dialog.currentDialog && (direction.x != 0 || direction.y != 0)) {
    if (baxton.footstepSfx == null) {
      //@ts-ignore
      baxton.footstepSfx = getSound(`baxton.footsteps`).play({ loop: true });
    }

    if (baxton.footstepFxTimer <= 0) {
      baxton.footstepFxTimer = 0.2;
      spawnAnimatedFx(`inGame/fxs/run`, baxton.pos, 0, `anim${randInt(1, 3)}`, inGame.activeWorld.pixiShadowsLayer);
    } else {
      baxton.footstepFxTimer -= dt;
    }

    const angle = Math.atan2(direction.y, direction.x);

    const move = new PIXI.Point(
      (Math.round(Math.cos(angle) * 1000) / 1000) * speed,
      (Math.round(Math.sin(angle) * 1000) / 1000) * speed
    );

    moveWithCollision(baxton, move);

    setAnimation(baxton.bodySprite, baxton.bodySpritesheet.animations.run);
    // offset arm while running
    centerToArmSocket.x += 5;
  } else {
    stopBaxtonFootsteps();

    baxton.footstepFxTimer = 0;


    baxton.footstepFxTimer = 0;

    setAnimation(baxton.bodySprite, baxton.bodySpritesheet.animations.idle);
  }

  let velocity = new PIXI.Point(
    baxton.pos.x - oldPos.x,
    baxton.pos.y - oldPos.y,
  );

  updateRenderPos();

  // Aiming
  const aimsLeft = target.pos.x < baxton.pos.x;

  baxton.armAngle = Math.atan2(target.pos.y - baxton.pos.y + 0.6, target.pos.x - baxton.pos.x);
  setBaxtonDirection(aimsLeft, centerToArmSocket);

  // Shooting
  if (baxton.shootCooldown > 0) {
    baxton.shootCooldown -= dt;
  }

  const weaponIsInCooldown = baxton.shootCooldown > 0;
  const tryingToFire = !inGame.hud.isCrafting && mouse.buttons[0] === true;

  if (tryingToFire && !inGame.dialog.currentDialog) {
    baxton.armSprite.texture = tex("inGame/baxton/baxtonArmFiring");
  } else {
    baxton.armSprite.texture = tex("inGame/baxton/baxtonArmIdle");
    baxton.armSprite.angle = 0;
  }

  if (tryingToFire && !weaponIsInCooldown && !inGame.dialog.currentDialog) {
    let ammoDesc = tryShoot();
    if (ammoDesc) {
      let xFactor = aimsLeft ? -1 : 1;
      let armSocketPos = new PIXI.Point(
        baxton.pos.x + centerToArmSocket.x * inverseTileSize * xFactor,
        baxton.pos.y + centerToArmSocket.y * inverseTileSize,
      );

      let socketAngle = Math.atan2(
        target.pos.y - armSocketPos.y,
        target.pos.x - armSocketPos.x,
      );

      let sth = playerMetrics.socketToHole;

      let yAngle = (aimsLeft ? -1 : 1) * Math.PI / 2;

      let canonPos = new PIXI.Point(
        armSocketPos.x + Math.cos(socketAngle) * sth.x * inverseTileSize + Math.cos(socketAngle + yAngle) * sth.y * inverseTileSize,
        armSocketPos.y + Math.sin(socketAngle) * sth.x * inverseTileSize + Math.sin(socketAngle + yAngle) * sth.y * inverseTileSize,
      );

      let bulletPos = new PIXI.Point();
      bulletPos.copyFrom(canonPos);

      let canonAngle = Math.atan2(
        target.pos.y - canonPos.y,
        target.pos.x - canonPos.x,
      );

      let badAimInnerRadius = 1;
      let badAimOuterRadius = 4;
      let gunDist = distxy(target.pos, canonPos);
      /** @type {number} */
      let badAimFactor;
      if (gunDist < badAimInnerRadius) {
        badAimFactor = 1;
      } else if (gunDist < badAimOuterRadius) {
        badAimFactor = (gunDist - badAimInnerRadius) / (badAimOuterRadius - badAimInnerRadius);
      } else {
        badAimFactor = 0;
      }

      let finalAngle = lerpAngle(canonAngle, socketAngle, badAimFactor);

      const bulletVelocity = new PIXI.Point(
        Math.cos(finalAngle) * ammoDesc.speed + velocity.x,
        Math.sin(finalAngle) * ammoDesc.speed + velocity.y,
      );
      spawnBullet(bulletPos, bulletVelocity, ammoDesc, "baddies");

      getSound(`shoot.${ammoDesc.id}`).play();
      baxton.shootCooldown = ammoDesc.cooldown;
    } else {
      if (mouse.presses[0]) getSound(`shoot.empty`).play();
    }
  }

  // note: AnimatedSprite.update needs ticks
  baxton.bodySprite.update(app.ticker.deltaTime);
}

function setBaxtonDirection(lookingLeft, centerToArmSocket) {
  const { baxton } = inGame;

  baxton.bodySprite.scale.x = lookingLeft ? -1 : 1;

  baxton.armContainer.position.set(
    (lookingLeft ? -1 : 1) * centerToArmSocket.x,
    centerToArmSocket.y,
  );

  baxton.armSprite.scale.x = lookingLeft ? -1 : 1;
  baxton.armSprite.angle = (lookingLeft ? baxton.armAngle - Math.PI : baxton.armAngle) * PIXI.RAD_TO_DEG;
}

/**
 * @param {number} damage
 * @param {number} angle
 */
function damageBaxton(damage, angle) {
  const { baxton } = inGame;
  if (baxton.immuneTimer > 0) return;

  if (inGame.stats.armor > 0) inGame.stats.armor = Math.max(inGame.stats.armor - damage, 0);
  else inGame.stats.health = Math.max(inGame.stats.health - damage, 0);
  updateHudHealth();

  const movingRight = angle > Math.PI / 2 || angle < -Math.PI / 2;
  setBaxtonDirection(movingRight, playerMetrics.centerToArmSocket);
  baxton.armSprite.texture = tex("inGame/baxton/baxtonArmIdle");
  baxton.armSprite.angle = 0;

  getSound(`${inGame.stats.health === 0 ? "deaths" : "hits"}.baxton`).play();

  if (inGame.stats.health === 0) {
    baxton.pixiRoot.removeChild(baxton.armContainer);

    baxton.bodySprite.loop = false;
    setAnimation(baxton.bodySprite, baxton.bodySpritesheet.animations.die);
    stopBaxtonFootsteps();

    setInGameOver();
  } else {
    baxton.hitKnockback.set(-Math.cos(angle) * 0.05, -Math.sin(angle) * 0.05);
    baxton.hitTimer = 0.3;

    baxton.immuneTimer = 1;

    baxton.bodySprite.loop = false;
    setAnimation(baxton.bodySprite, baxton.bodySpritesheet.animations.hit);
  }
}

function setupBaxtonArea(ignoreIfNone) {
  const { baxton } = inGame;

  for (const area of inGame.activeWorld.areas) {
    if (area.rect.contains(baxton.pos.x, baxton.pos.y)) {
      baxton.area = area;
      ensureInGameMusic(`worlds.${inGame.activeWorld.name}.${area.name}`);
      ensureWeather(area.weather);
      return;
    }
  }

  if (!ignoreIfNone) {
    ensureInGameMusic(`worlds.${inGame.activeWorld.name}.main`);
    ensureWeather(null);
  }
}

function stopBaxtonFootsteps() {
  if (inGame.baxton.footstepSfx != null) {
    inGame.baxton.footstepSfx.stop();
    inGame.baxton.footstepSfx = null;
  }
}